/* $Id$ */
#include <X11/keysym.h> /* XK_q */
#include <X11/Xutil.h> /* XDestroyImage() */
#include "common.h"
#include "cache.h"
#include "dboard.h"
#include "loader.h"
#include "basetime.h"
#include "option.h"
#include "image_utils.h"
#include "cmsprofile.h"

#define TEMPERATURE 6500
#define SCREEN_GAMMA 2.2

struct xImageWork {
  int dispose_op;
  XImage *previous;
  int width;
  int height;
};
typedef struct xImageWork XImageWork;

struct wrapper {
  unsigned int index;
  DBoard *board;
  ImageWork *work;
  XImageWork *xWork;
};

struct cachedImage {
  XImage *ximage;
  int xoff;
  int yoff;
  int width;
  int height;
};

struct player {
  ImageSet *images;
  Image *current;
  Cache *cache;
  int disable_cache;
  ImageWork *imageWork;
  XImageWork xImageWork;
  int count;
  int loop_count;
  int running, running0;
  BaseTime *t0, *t;
  const char *format;
  int argc;
  const char **argv;
  int arg_index, arg_index0;
  const char *icc_input_file;
  const char *icc_output_file;
  int icc_input_use;
  int icc_input_sRGB;
  int icc_output_sRGB;
  CmsProfile *profile_input;
  CmsProfile *profile_output;
  CmsProfile *profile_sRGB;
};
typedef struct player Player;

static int x_main(int argc, char **argv);

static void player_init(Player *player, int argc, const char **argv, int arg_index);
static void player_free(Player *player);
static void player_image_next(Player *player);
static void player_arg_next(Player *player);
static void player_arg_previous(Player *player);
static int player_read_file(Player *player);
static void player_reset(Player *player);
static int player_has_multiframe(Player *player);

static int moving_process(DBoard *board, int *pdone, void *data);
static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data);
static void board_setup(DBoard *board, Player *player);
static void image_setup(DBoard *board, Player *player);

static unsigned int ximage_hashcode(const void *arg);
static void *ximage_get(const void *arg);
static int ximage_release(void *data);

static XImage *ximage_setup(DBoard *board, ImageWork *work, XImageWork *xWork);
static void ximage_indexed(XImage *ximage, ImageWork *work, XColor screen_color[][SCREEN_RGB_IMAX]);
static void ximage_mapped(XImage *ximage, ImageWork *work, XColor screen_color[][SCREEN_RGB_IMAX], unsigned long (*xcolor_get_pixel)(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char *pp));
static void ximage_pixel(XImage *ximage, ImageWork *work, DBoard *board, unsigned long (*dBoard_get_pixel)(DBoard *board, unsigned char *pp));
static unsigned long xcolor_background_pixel(DBoard *board, ImageWork *work);
static unsigned long xcolor_get_pixel_rgb8(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char *p);
static unsigned long xcolor_get_pixel(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char r, unsigned char g, unsigned char b);

static void dBoard_fill_rectangle(DBoard *board, int x, int y, unsigned int width, unsigned int height, unsigned long pixel);
static unsigned long dBoard_get_pixel_rgb16(DBoard *board, unsigned char *p);
static unsigned long dBoard_get_colorpixel(DBoard *board, unsigned short r, unsigned short g, unsigned short b);


int main(int argc, char **argv)
{
  x_main(argc, argv);
  return EXIT_SUCCESS;
}

static int x_main(int argc, char **argv)
{
  DBoard *board;
  Player player;
  OptionEntry option[] = {
    { "animate", "a", OPTION_ARG_NONE, &(player.running0) },
    { "format", NULL, OPTION_ARG_STRING, &(player.format) },
    { "disable-cache", NULL, OPTION_ARG_NONE, &(player.disable_cache) },
    { "input-no-embedded", "in", OPTION_ARG_NONE, &(player.icc_input_use) },
    { "input-profile", "ip", OPTION_ARG_STRING, &(player.icc_input_file) },
    { "input-sRGB", "is", OPTION_ARG_NONE, &(player.icc_input_sRGB) },
    { "output-profile", "op", OPTION_ARG_STRING, &(player.icc_output_file) },
    { "output-sRGB", "os", OPTION_ARG_NONE, &(player.icc_output_sRGB) },
    { NULL, NULL, OPTION_ARG_NONE, NULL }
  };

  board = dBoard_open("xiview", &argc, argv);
  if (board == NULL) {
    fprintf(stderr, "Cannot open toolkit.\n");
    exit(EXIT_FAILURE);
  }
  player.running0 = FALSE;
  player.format = NULL;
  player.disable_cache = FALSE;
  player.icc_input_use = FALSE;
  player.icc_input_file = NULL;
  player.icc_input_sRGB = FALSE;
  player.icc_output_file = NULL;
  player.icc_output_sRGB = FALSE;
  argc--;
  argv++;
  if (!option_get(&argc, &argv, option)) {
    /* usage */
    exit(EXIT_FAILURE);
  }
  player_init(&player, argc, argv, 0);

  if (!player_read_file(&player)) {
    fprintf(stderr, "Cannot read file.\n");
    player_free(&player);
    exit(EXIT_FAILURE);
  }
  board_setup(board, &player);
  image_setup(board, &player);
  dBoard_setHandler(board, handle_event,
                    ExposureMask |
                    KeyPressMask | KeyReleaseMask | ColormapChangeMask);
  dBoard_setProcessor(board, moving_process);
  dBoard_show(board);
  while (!dBoard_main(board, &player)) {
    ;
  }
  dBoard_close(board);
  player_free(&player);
  return TRUE;
}

static void player_init(Player *player, int argc, const char **argv, int arg_index)
{
  player->t0 = baseTime_new();
  player->t = baseTime_new();
  player->argc = argc;
  player->argv = argv;
  player->arg_index = player->arg_index0 = arg_index;
  player->images = imageSet_new();
  player->imageWork = imageWork_alloc();
  player->cache = NULL;
  player->profile_input = NULL;
  if (player->icc_input_file != NULL) {
    player->profile_input = cmsProfile_open_file(player->icc_input_file);
  }
  player->profile_output = NULL;
  if (player->icc_output_file != NULL) {
    player->profile_output = cmsProfile_open_file(player->icc_output_file);
  } else {
#if 0
    int tempK = TEMPERATURE;
    double gamma[3];
    for (int i = 0; i < 3; i++) {
      gamma[i] = 1.0/SCREEN_GAMMA;
    }
    player->profile_output = cmsProfile_open_temp(tempK, gamma, 3);
#endif /* 0 */
    player->profile_output = NULL;
  }
  player->profile_sRGB = cmsProfile_open_sRGB();
}

static void player_free(Player *player)
{
  baseTime_delete(player->t0);
  baseTime_delete(player->t);
  imageSet_free(player->images);
  XFREE(player->images);
  imageWork_free(player->imageWork);
  cache_close(player->cache);
  cmsProfile_close(player->profile_input);
  cmsProfile_close(player->profile_output);
  cmsProfile_close(player->profile_sRGB);
}

static void player_image_next(Player *player)
{
  Image *image;
  int num_loop;

  image = player->current;
  if (image_next(image) == NULL) {
    player->current = imageSet_get_first(player->images);
    player->count = 0;
    player->loop_count++;
  } else {
    player->current = image_next(image);
    player->count++;
  }
  num_loop = imageSet_get_num_plays(player->images);
  if (num_loop > 0) {
    if (player->loop_count >= num_loop) {
      player->running = FALSE;
    }
  }
}

static void player_arg_next(Player *player)
{
  if (++(player->arg_index) >= player->argc) {
    player->arg_index = player->argc-1;
  }
}

static void player_arg_previous(Player *player)
{
  if(--(player->arg_index) < 1) {
    player->arg_index = player->arg_index0;
  }
}

static int player_read_file(Player *player)
{
  Entry entry;
  const char *filename;
  const char *format;
  int index, rc;

  index = player->arg_index;
  filename = (player->argv)[index];
  format = player->format;
  DEBUG_PRINT("filename = '%s'\n", filename);
  entry.filename = filename;
  cache_close(player->cache);
  imageSet_free(player->images);
  rc = load_images(&entry, player->images, format);
  player_reset(player);
  if (!rc) {
    return FALSE;
  }
  return TRUE;
}

static void player_reset(Player *player)
{
  player->current = imageSet_get_first(player->images);
  player->count = 0;
  player->loop_count = 0;
  if (player_has_multiframe(player)) {
    player->running = player->running0;
    if (!player->disable_cache) {
      player->cache = cache_open(imageSet_get_num_frames(player->images), ximage_hashcode, ximage_get, ximage_release);
    } else {
      player->cache = NULL;
    }
  } else {
    player->running = FALSE;
    player->cache = NULL;
  }
  baseTime_get(player->t0);
}

static int player_has_multiframe(Player *player)
{
  return (imageSet_get_num_frames(player->images) > 1);
}

static int moving_process(DBoard *board, int *pdone, void *data)
{
  Player *player;
  BaseTime *t;
  int dt, delay;

  player = (Player *)data;
  if (!player->running) {
    baseTime_sleep_msec(20);
    return TRUE;
  }
  delay = image_get_delay(player->current);
  t = player->t;
  baseTime_get(t);
  dt = baseTime_subtract_msec(t, player->t0);
  if (dt >= delay) {
    player_image_next(player);
    image_setup(board, player);
    dBoard_copy(board);
    baseTime_get(t);
    dt = baseTime_subtract_msec(t, player->t0);
    baseTime_set(player->t0, t);
    DEBUG_PRINT("dt = %d, delay = %d\n", dt, delay);
  } else {
    int wait = (delay-dt)/10;
    if (wait < 20) {
      baseTime_sleep_msec(wait);
    } else {
      baseTime_sleep_msec(20);
    }
  }
  return TRUE;
}

static int handle_event(DBoard *board, XEvent *event, int *pdone, void *data)
{
  Player *player;
  int rc, done;
  Atom delwin;
  KeySym keysym;

  player = (Player *)data;
  rc = done = FALSE;
  switch (event->type) {
  case Expose:
    DEBUG_PRINT("Expose\n");
    if ((event->xexpose).count == 0 ) {
      dBoard_copy(board);
      dBoard_flush(board);
    }
    break;
  case NoExpose:
    DEBUG_PRINT("NoExpose\n");
    break;
  case ClientMessage:
    DEBUG_PRINT("ClientMessage\n");
    delwin = XInternAtom(board->display, "WM_DELETE_WINDOW", True);
    if ((event->xclient).data.l[0] == delwin) {
      done = TRUE;
    }
    break;
  case KeyRelease:
    keysym = XLookupKeysym(&(event->xkey), 0);
    if (keysym == XK_n) {
      if (player_has_multiframe(player)) {
        player_image_next(player);
        image_setup(board, player);
        dBoard_copy(board);
        dBoard_flush(board);
      }
    } else if (keysym == XK_space) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_next(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_BackSpace) {
      dBoard_define_cursor(board, CURSOR_WAIT);
      dBoard_flush(board);
      player_arg_previous(player);
      player_read_file(player);
      board_setup(board, player);
      image_setup(board, player);
      dBoard_copy(board);
      dBoard_flush(board);
      dBoard_undefine_cursor(board);
    } else if (keysym == XK_s) {
      if (player_has_multiframe(player)) {
        player->running = !(player->running);
      }
    } else if (keysym == XK_q) {
      done = TRUE;
    } else {
      ;
    }
    break;
  case DestroyNotify:
    DEBUG_PRINT("DestroyNotify\n");
    done = TRUE;
    break;
  default:
    DEBUG_PRINT("event: %d\n", event->type);
    break;
  }
  *pdone = done;
  return rc;
}

static void board_setup(DBoard *board, Player *player)
{
  int width, height;
  int index;
  const char *filename;

  imageSet_get_size(player->images, &width, &height);
  dBoard_reset(board, width, height);
  index = player->arg_index;
  filename = (player->argv)[index];
  dBoard_set_name(board, filename);

  BgPrefer bg_prefer;
  unsigned short bg_r, bg_g, bg_b;
  unsigned char bgdata[6];
  int bg_setted;
  int depth, w, h;
  int use_input;
  CmsProfile *profile_in;
  CmsProfile *profile_out;

  use_input = player->icc_input_use;
  profile_in = (player->icc_input_sRGB) ? player->profile_sRGB : player->profile_input;
  profile_out = (player->icc_output_sRGB) ? player->profile_sRGB : player->profile_output;
  bg_r = bg_g = bg_b = 0x7f7f;
  dBoard_get_attributes(board, &depth, &w, &h, &bg_r, &bg_g, &bg_b, &bg_setted);
  DEBUG_PRINT(" attr: %d [%dx%d] bg_setted=%s (%x,%x,%x)\n",
              depth, w, h, (bg_setted) ? "TRUE" : "FALSE", bg_r, bg_g, bg_b);
  imageWork_init(player->imageWork, player->images);
  imageWork_set_profiles(player->imageWork, profile_in, profile_out, use_input);
  bg_prefer = (bg_setted) ? BG_USER : BG_IMAGE;

  bgdata[0] = (bg_r >> 8) & 0xff;
  bgdata[1] = bg_r & 0xff;
  bgdata[2] = (bg_g >> 8) & 0xff;
  bgdata[3] = bg_g & 0xff;
  bgdata[4] = (bg_b >> 8) & 0xff;
  bgdata[5] = bg_b & 0xff;
  imageWork_set_background(player->imageWork, bgdata, bg_prefer);
}

static void image_setup(DBoard *board, Player *player)
{
  Display *display;
  GC gc;
  Pixmap pixmap;
  XImage *ximage;
  XImage *last_frame;
  XImageWork *xWork;
  ImageWork *imageWork;
  Image *image;
  int width, height;
  int xoff, yoff;
  int dispose_op;

  image = player->current;
  if (image == NULL) {
    return;
  }
  imageWork = player->imageWork;
  xWork = &(player->xImageWork);
  display = board->display;
  pixmap = board->pixmap;
  if (player->cache != NULL) {
    struct wrapper t;
    struct cachedImage *cached;

    t.index = player->count;
    cached = (struct cachedImage *)cache_tryget(player->cache, &t);
    if (cached == NULL) {
      int w, h;
      last_frame = NULL;
      dispose_op = image_get_dispose_op(image);
      DEBUG_PRINT(" dispose_op = %d\n", dispose_op);
      switch (dispose_op) {
      case DISPOSE_OP_PREVIOUS:
        imageSet_get_size(player->images, &w, &h);
        last_frame = XGetImage(display, pixmap, 0, 0, w, h, AllPlanes, ZPixmap);
        break;
      }
      imageWork_update(imageWork, image);
      t.board = board;
      t.work = imageWork;
      t.xWork = xWork;
      cached = (struct cachedImage *)cache_get(player->cache, &t);
    }
    xoff = cached->xoff;
    yoff = cached->yoff;
    width = cached->width;
    height = cached->height;
    ximage = cached->ximage;
  } else {
    int w, h;
    last_frame = NULL;
    dispose_op = image_get_dispose_op(image);
    DEBUG_PRINT(" dispose_op = %d\n", dispose_op);
    switch (dispose_op) {
    case DISPOSE_OP_PREVIOUS:
      imageSet_get_size(player->images, &w, &h);
      last_frame = XGetImage(display, pixmap, 0, 0, w, h, AllPlanes, ZPixmap);
      break;
    }
    imageWork_update(imageWork, image);
    imageWork_get_updated_size(imageWork, &width, &height, &xoff, &yoff);
    ximage = ximage_setup(board, imageWork, xWork);
  }
  xWork->dispose_op = dispose_op;
  xWork->previous = last_frame;
  DEBUG_PRINT(" updated image : (%d, %d) [%dx%d]\n",
              xoff, yoff, width, height);
  gc = board->gc;
  XPutImage(display, pixmap, gc, ximage, 0, 0, xoff, yoff, width, height);
  if (player->cache == NULL) {
    XDestroyImage(ximage);
  }
}

static unsigned int ximage_hashcode(const void *arg)
{
  struct wrapper *p;

  p = (struct wrapper *)arg;
  return p->index;
}

static void *ximage_get(const void *arg)
{
  struct wrapper *p;
  struct cachedImage *cached;
  int xoff;
  int yoff;
  int width;
  int height;

  p = (struct wrapper *)arg;
  imageWork_get_updated_size(p->work, &width, &height, &xoff, &yoff);
  cached = XMALLOC(struct cachedImage, 1);
  cached->xoff = xoff;
  cached->yoff = yoff;
  cached->width = width;
  cached->height = height;
  cached->ximage = ximage_setup(p->board, p->work, p->xWork);
  return cached;
}

static int ximage_release(void *p)
{
  struct cachedImage *cached;
  int rc;

  cached = (struct cachedImage *)p;
  rc = XDestroyImage(cached->ximage);
  XFREE(cached);
  return rc;
}

static XImage *ximage_setup(DBoard *board, ImageWork *work, XImageWork *xWork)
{
  Display *display;
  Pixmap pixmap;
  GC gc;
  XImage *ximage;
  int dispose_op, blend_op;
  int width, height;
  int xoff, yoff;
  ImageType type;
  Image *image;
  unsigned long background;
  XImage *prev_ximage;

  display = board->display;
  pixmap = board->pixmap;
  gc = board->gc;

  image = imageWork_get_image(work);
  type = image_get_type(image);
  imageWork_get_updated_size(work, &width, &height, &xoff, &yoff);
  /* dispose image */
  dispose_op = imageWork_get_dispose_op(work);
  DEBUG_PRINT(" dispose_op type: %d\n", dispose_op);
  switch (dispose_op) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    background = xcolor_background_pixel(board, work);
    dBoard_fill_rectangle(board, xoff, yoff, width, height, background);
    break;
  case DISPOSE_OP_PREVIOUS:
    XPutImage(display, pixmap, gc, xWork->previous, 0, 0, xoff, yoff, width, height);
    XDestroyImage(xWork->previous);
    break;
  default:
    DEBUG_PRINT(" unknown dipose_op type %d. why?\n", dispose_op);
    break;
  }
  blend_op = image_get_blend_op(image);
  DEBUG_PRINT(" blend_op type: %d\n", blend_op);
  switch (blend_op) {
  case BLEND_OP_NONE:
    ximage = XGetImage(display, pixmap, xoff, yoff, width, height, AllPlanes, ZPixmap);
    break;
  case BLEND_OP_SOURCE:
    background = xcolor_background_pixel(board, work);
    dBoard_fill_rectangle(board, xoff, yoff, width, height, background);
    ximage = XGetImage(display, pixmap, xoff, yoff, width, height, AllPlanes, ZPixmap);
    break;
  case BLEND_OP_OVER:
    ximage = XGetImage(display, pixmap, xoff, yoff, width, height, AllPlanes, ZPixmap);
    break;
  default:
    DEBUG_PRINT(" unknown blend_op type %d. why?\n", blend_op);
    break;
  }
  DEBUG_PRINT(" image type: %d %d (%d, %d) [%dx%d]\n",
              type, image_get_num_indexes(image),
              xoff, yoff, width, height);
  switch (type) {
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    DEBUG_PRINT("image_type_indexed. type=%d\n", type);
    ximage_indexed(ximage, work, board->screen_color);
    break;
  case IMAGE_TYPE_RGB_8:
    DEBUG_PRINT("image_type_rgb8.\n");
    ximage_mapped(ximage, work, board->screen_color, xcolor_get_pixel_rgb8);
    break;
  case IMAGE_TYPE_RGB_16:
    DEBUG_PRINT("image_type_rgb16.\n");
    ximage_pixel(ximage, work, board, dBoard_get_pixel_rgb16);
    break;
  default:
    fprintf(stderr, "unknown image type %d. why?\n", type);
    break;
  }
  return ximage;
}

static void ximage_indexed(XImage *ximage, ImageWork *work, XColor screen_color[][SCREEN_RGB_IMAX])
{
  unsigned long xcolors[SCREEN_RGB_IMAX];
  unsigned char alpha[SCREEN_RGB_IMAX];
  ImageType type;
  Image *image;
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;
  int num;

  image = imageWork_get_image(work);
  type = image_get_type(image);
  bpp = imageType_bytes_per_pixel(type);
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);
  num = image_get_num_indexes(image);
  y = yoff;
  switch (type) {
  case IMAGE_TYPE_INDEXED_RGB_8:
    for (int i = 0; i < num; i++) {
      unsigned char data[3];
      image_get_index(image, i, data);
      xcolors[i] = xcolor_get_pixel(screen_color, data[0], data[1], data[2]);
    }
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp;
      for (int i = 0; i < w; i++) {
        int c;
        c = *(pp++);
        XPutPixel(ximage, i, j, xcolors[c]);
      }
      y++;
    }
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    for (int i = 0; i < num; i++) {
      unsigned char data[4];
      image_get_index(image, i, data);
      xcolors[i] = xcolor_get_pixel(screen_color, data[0], data[1], data[2]);
      alpha[i] = data[3];
    }
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp;
      for (int i = 0; i < w; i++) {
        int c;
        c = *(pp++);
        if (alpha[c] >= 0x80) {
          /* fully opaque */
          XPutPixel(ximage, i, j, xcolors[c]);
        }
      }
      y++;
    }
    break;
  }
}

static void ximage_mapped(XImage *ximage, ImageWork *work, XColor screen_color[][SCREEN_RGB_IMAX], unsigned long (*xcolor_get_pixel)(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char *pp))
{
  Image *image;
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  image = imageWork_get_image(work);
  bpp = imageType_bytes_per_pixel(image_get_type(image));
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);
  y = yoff;
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*y+xoff)*bpp;
    for (int i = 0; i < w; i++) {
      unsigned long xcol;
      xcol = (*xcolor_get_pixel)(screen_color, pp);
      XPutPixel(ximage, i, j, xcol);
      pp += bpp;
    }
    y++;
  }
}

static void ximage_pixel(XImage *ximage, ImageWork *work, DBoard *board, unsigned long (*dBoard_get_pixel)(DBoard *board, unsigned char *pp))
{
  Image *image;
  unsigned char *pp1;
  int bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;

  image = imageWork_get_image(work);
  bpp = imageType_bytes_per_pixel(image_get_type(image));
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  imageWork_get_updated_size(work, &w, &h, &xoff, &yoff);
  y = yoff;
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*y+xoff)*bpp;
    for (int i = 0; i < w; i++) {
      unsigned long xcol;
      xcol = (*dBoard_get_pixel)(board, pp);
      XPutPixel(ximage, i, j, xcol);
      pp += bpp;
    }
    y++;
  }
}
#if 0
static void xput_rgb8(DBoard *board,
                      XImage ximage, int width, int height,
                      const unsigned char *pixel)
{
  unsigned long xcol;
  xcol = xcolor_get_pixel_rgb8(board->screen_color, pixel);
  XPutPixel(ximage, i, j, xcol);
}
static void xput_rgb16(DBoard *board,
                       XImage ximage, int width, int height,
                       const unsigned char *pixel)
{
  unsigned long xcol;
  xcol = dBoard_get_pixel_rgb16(board, pixel);
  XPutPixel(ximage, i, j, xcol);
}
#endif /* 0 */

static unsigned long xcolor_background_pixel(DBoard *board, ImageWork *work)
{
  const unsigned char unknown[] = { 0x7f, 0x7f, 0x7f };
  unsigned long pixel;
  unsigned char data[8];
  ImageType type;

  imageWork_get_background(work, data, &type);
  switch (type) {
  case IMAGE_TYPE_RGB_8:
    pixel = xcolor_get_pixel_rgb8(board->screen_color, data);
    break;
  case IMAGE_TYPE_RGB_16:
    pixel = dBoard_get_pixel_rgb16(board, data);
    break;
  case IMAGE_TYPE_UNKNOWN:
    pixel = xcolor_get_pixel_rgb8(board->screen_color, unknown);
    DEBUG_PRINT(" background type = unknown\n");
    break;
  default:
    DEBUG_PRINT(" background type = %d, why?\n", type);
    break;
  }
  DEBUG_PRINT(" pixel = %x\n", pixel);
  return pixel;
}

static unsigned long xcolor_get_pixel_rgb8(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char *p)
{
  unsigned char r, g, b;
  r = *p++;
  g = *p++;
  b = *p++;
  return xcolor_get_pixel(screen_color, r, g, b);
}

static unsigned long xcolor_get_pixel(XColor screen_color[][SCREEN_RGB_IMAX], unsigned char r, unsigned char g, unsigned char b)
{
  unsigned long pixel;
  pixel  = screen_color[0][r].pixel;
  pixel |= screen_color[1][g].pixel;
  pixel |= screen_color[2][b].pixel;
  return pixel;
}

static void dBoard_fill_rectangle(DBoard *board, int x, int y, unsigned int width, unsigned int height, unsigned long pixel)
{
  Display *display;
  Pixmap pixmap;
  GC gc;

  display = board->display;
  pixmap = board->pixmap;
  gc = board->gc;
  XSetForeground(display, gc, pixel);
  XFillRectangle(display, pixmap, gc, x, y, width, height);
}

static unsigned long dBoard_get_pixel_rgb16(DBoard *board, unsigned char *p)
{
  unsigned short c16[3];

  for (int i = 0; i < 3; i++) {
    c16[i]  = (*(p++) << 8) & 0xff00;
    c16[i] |= *(p++) & 0xff;
  }
  return dBoard_get_colorpixel(board, c16[0], c16[1], c16[2]);
}

static unsigned long dBoard_get_colorpixel(DBoard *board, unsigned short r, unsigned short g, unsigned short b)
{
  Display *display;
  int screen;
  Colormap cmap;
  XColor color;

  display = board->display;
  screen = XDefaultScreen(display);
  cmap = XDefaultColormap(display, screen);
  color.red = r;
  color.green = g;
  color.blue = b;
  XAllocColor(display, cmap, &color);
  return color.pixel;
}
